import copy
import warnings

import gaussianprocessderivatives as gp
import matplotlib.pylab as plt
import numpy as np
from scipy.interpolate import interp1d

import omniplate.omgenutils as gu


class fitderiv:
    """
    Smooth and estimate the time derivative of the data via Gaussian processes.

    After a successful optimisation, the following attributes are generated:

    t: array
        The times specified as input.
    d: array
        The data specified as input.
    f: array
        The mean of the Gaussian process with the optimal hyperparmeters
        at each time point.
    fvar: array
        The variance of the optimal Gaussian process at each time point.
    df: array
        The inferred first time-derivative.
    dfvar: array
        The inferred variance of the first time-derivative.
    ddf: array
        The inferred second time-derivative.
    ddfvar: array
        The inferred variance of the second time-derivative.

    Examples
    --------
    A typical work flow is:

    >>> from fitderiv import fitderiv
    >>> q= fitderiv(t, od, figs= True)
    >>> q.plotfit('df')

    or potentially

    >>> import matplotlib.pyplot as plt
    >>> plt.plot(q.t, q.d, 'r.', q.t, q.y, 'b')

    Reference
    ---------
    PS Swain, K Stevenson, A Leary, LF Montano-Gutierrez, IBN Clark, J Vogel,
    and T Pilizota.
    Inferring time derivatives including growth rates using Gaussian processes
    Nat Commun 7 (2016) 13766
    """

    def __init__(
        self,
        t,
        d,
        cvfn="matern",
        logs=True,
        noruns=5,
        noinits=100,
        exitearly=False,
        bd=False,
        empirical_errors=False,
        optmethod="L-BFGS-B",
        warn=False,
        linalgmax=3,
        max_data_pts=None,
    ):
        """
        Smooth data and estimate time derivatives with a Gaussian process.

        Parameters
        ----------
        t: 1D array
            The time points.
        d: array
            The data corresponding to the time points with any replicates given
             as columns.
        cvfn: string
            The type of kernel function for the Gaussian process either 'sqexp'
            (squared exponential) or 'matern' (Matern with nu= 5/2) or 'nn'
            (neural network).
        logs: boolean
            If True, the Gaussian process is used to smooth the natural
            logarithm of the data and the time-derivative is therefore of the
            logarithm of the data.
        noruns: integer, optional
            The number of attempts to be made at optimising the kernel's
            hyperparmeters.
        noinits: integer, optional
            The number of random attempts made to find good initial choices for
            the hyperparameters before running their optimisation.
        exitearly: boolean, optional
            If True, stop at the first successful attempt at optimising the
            hyperparameters otherwise take the best choice from all successful
            optimisations.
        bd: dictionary, optional
            Specifies the limits on the hyperparameters for the Gaussian process.
            For example, bd= {0: [-1, 4], 2: [2, 6]})
            sets confines the first hyperparameter to be between 1e-1 and 1e^4
            and confines the third hyperparmater between 1e2 and 1e6.
        empirical_errors: boolean, optional
            If True, measurement errors are empirically estimated by the
            variance across replicates at each time point.
            If False, the variance of the measurement error is assumed to be
            the same for all time points and its magnitude is a hyperparameter
            that is optimised.
        optmethod: string, optional
            The algorithm used to optimise the hyperparameters, either
            'l_bfgs_b' or 'tnc'.
        warn: boolean, optional
            If False, warnings created by covariance matrices that are not
            positive semi-definite are suppressed.
        linalgmax: integer, optional
            The number of times errors generated by underlying linear algebra
            modules during the optimisation by poor choices of the
            hyperparameters should be ignored.
        max_data_pts: integer, optional
            If set, sufficiently large data sets with multiple replicates will
            be subsampled at each time point, randomly picking a smaller
            number of replicates, to reduce the number of data points and so
            run times.
        """
        self.linalgmax = linalgmax
        self.success = True
        if not warn:
            # warning generated occasionally when sampling from the Gaussian
            # process likely because of numerical errors
            warnings.simplefilter("ignore", RuntimeWarning)
        try:
            noreps = d.shape[1]
        except IndexError:
            noreps = 1
        self.noreps = noreps
        self.d = np.copy(d)
        self.t = np.copy(t)
        t = self.t
        d = self.d
        # default bounds for hyperparameters
        bddict = {
            "nn": {0: (-1, 5), 1: (-7, -2), 2: (-6, 2)},
            "sqexp": {0: (-5, 5), 1: (-6, 2), 2: (-5, 2)},
            "matern": {0: (-5, 5), 1: (-4, 4), 2: (-5, 2)},
        }
        # find bounds
        if bd:
            self.bds = gu.mergedicts(original=bddict[cvfn], update=bd)
        else:
            self.bds = bddict[cvfn]
        # log data
        self.logs = logs
        if logs:
            print("Taking natural logarithm of the data.")
            if np.any(np.nonzero(d < 0)):
                print("Warning: Negative data is being set to NaN.")
            # replace zeros and negs so that logs can be applied
            d[np.nonzero(d <= 0)] = np.nan
            # take log of data
            d = np.log(np.asarray(d))
        # run checks and define measurement errors
        if empirical_errors:
            # errors must be empirically estimated
            if noreps > 1:
                lod = [
                    np.count_nonzero(np.isnan(d[:, i])) for i in range(noreps)
                ]
                if np.sum(np.diff(lod)) != 0:
                    print(
                        "The replicates have different number of data "
                        "points, but equal numbers of data points are "
                        "needed for empirically estimating errors."
                    )
                    merrors = None
                else:
                    # estimate errors empirically
                    print("Estimating measurement errors empirically.")
                    merrors = gu.findsmoothvariance(d)
            else:
                print("Not enough replicates to estimate errors empirically.")
                merrors = None
        else:
            merrors = None
        self.merrors = merrors
        # subsample if excessive data
        if max_data_pts and d[~np.isnan(d)].size > max_data_pts:
            print(
                f"Many data points - {d[~np.isnan(d)].size}:"
                " subsampling for each time point..."
            )
            no_samples_per_time_pt = np.ceil(max_data_pts / d.shape[0]).astype(
                "int"
            )
            d = subsample_data(d, no_samples_per_time_pt)
        # combine data into one array
        tb = np.tile(t, noreps)
        db = np.reshape(d, d.size, order="F")
        # check for NaNs
        if np.any(merrors):
            mb = np.tile(merrors, noreps)
            keep = np.intersect1d(
                np.nonzero(~np.isnan(db))[0], np.nonzero(~np.isnan(mb))[0]
            )
        else:
            keep = np.nonzero(~np.isnan(db))[0]
        # remove any NaNs
        da = db[keep]
        ta = tb[keep]
        # check data remains after removing NaNs
        if not da.size:
            print("Warning: omfitderiv failed - too many NaNs.")
            self.success = False
        elif np.any(merrors):
            # measurement errors
            ma = mb[keep]
            if not ma.size:
                print("Warning: omfitderiv failed - too many NaNs.")
                self.success = False
        else:
            ma = None
        if self.success:
            self.run(
                cvfn,
                ta,
                da,
                ma,
                noruns,
                noinits,
                exitearly,
                optmethod,
            )

    def run(
        self,
        cvfn,
        ta,
        da,
        ma,
        noruns,
        noinits,
        exitearly,
        optmethod,
    ):
        """Instantiate and run a Gaussian process."""
        try:
            # instantiate GP
            g = getattr(gp, cvfn + "GP")(self.bds, ta, da, merrors=ma)
            print("Using a " + g.description + ".")
            g.info
        except NameError:
            raise Exception("Gaussian process not recognised.")
        # optimise parameters
        g.findhyperparameters(
            noruns,
            noinits=noinits,
            exitearly=exitearly,
            optmethod=optmethod,
            linalgmax=self.linalgmax,
        )
        # display results
        g.results()
        if np.any(ma):
            # check measurement errors
            if len(ma) != len(self.t):
                # NaNs have been removed
                mainterp = interp1d(
                    ta, ma, bounds_error=False, fill_value=(ma[0], ma[-1])
                )
                ma = mainterp(self.t)
        g.predict(self.t, derivs=2, addnoise=True, merrorsnew=ma)
        # results
        self.g = g
        self.logmaxlike = -g.nlml_opt
        self.hparamerr = g.hparamerr
        self.lth = g.lth_opt
        self.f = g.f
        self.df = g.df
        self.ddf = g.ddf
        self.fvar = g.fvar
        self.dfvar = g.dfvar
        self.ddfvar = g.ddfvar

    def fitderivsample(self, nosamples, newt=None):
        """
        Generate samples from the latent function.

        Both values for the latent function and its first two
        derivatives are returned, as a tuple.

        All derivatives must be sampled because by default all are asked
        to be predicted by the underlying Gaussian process.

        Parameters
        ----------
        nosamples: integer
            The number of samples.
        newt: array, optional
            Time points for which the samples should be made.
            If None, the orginal time points are used.

        Returns
        -------
        samples: a tuple of arrays
            The first element of the tuple gives samples of the latent
            function;
            the second element gives samples of the first time derivative; and
            the third element gives samples of the second time derivative.
        """
        if np.any(newt):
            newt = np.asarray(newt)
            # make prediction for new time points
            gps = copy.deepcopy(self.g)
            gps.predict(newt, derivs=2, addnoise=True)
        else:
            gps = self.g
        samples = gps.sample(nosamples, derivs=2)
        return samples

    def plotfit(
        self, char="f", errorfac=1, xlabel="time", ylabel=False, figtitle=False
    ):
        """
        Plot the results of the fitting.

        Either the data and the mean of the optimal Gaussian process or
        the inferred time derivatives are plotted.

        Parameters
        ----------
        char: string
            The variable to plot either 'f' or 'df' or 'ddf'.
        errorfac: float, optional
            The size of the errorbars are errorfac times the standard deviation
            of the optimal Gaussian process.
        ylabel: string, optional
            A label for the y-axis.
        figtitle: string, optional
            A title for the figure.
        """
        x = getattr(self, char)
        xv = getattr(self, char + "var")
        if char == "f":
            d = np.log(self.d) if self.logs else self.d
            plt.plot(self.t, d, "r.")
        plt.plot(self.t, x, "b")
        plt.fill_between(
            self.t,
            x - errorfac * np.sqrt(xv),
            x + errorfac * np.sqrt(xv),
            facecolor="blue",
            alpha=0.2,
        )
        if ylabel:
            plt.ylabel(ylabel)
        else:
            plt.ylabel(char)
        plt.xlabel(xlabel)
        if figtitle:
            plt.title(figtitle)


def subsample_data(d, no_samples_per_time_pt):
    """
    Subsample replicate data.

    At each time point, randomly choose only some of the replicates.
    """
    if d.shape[1] > no_samples_per_time_pt:
        rng = np.random.default_rng()
        nd = rng.choice(d, no_samples_per_time_pt, axis=1, replace=False)
        return nd
    else:
        print(
            f"Number of samples, {no_samples_per_time_pt}, is "
            f"more than or equal to the number of replicates, {d.shape[1]}."
        )
        print("Subsampling stopped.")
        return d



# A Bayesian version of the method in Lichten ... Swain, BMC Biophys 2014
# TODO
# why is the predicted fluorescence so low? refstrain at wrong OD?
# delete code on priors - no longer used; general tidy
import numpy as np
from scipy.optimize import minimize
from tqdm import tqdm

import omniplate.corrections as omcorr
import omniplate.omgenutils as gu
import omniplate.sunder as sunder
from omniplate.runfitderiv import runfitderiv

rng = np.random.default_rng()

# notation follows Lichten ... Swain
# GFP is denoted y; AutoFL is denoted z.
# The reference strain is denoted WT.


def de_nan(y, z):
    """Remove any replicates with NaN."""
    # NaNs are generated because experiments have different durations
    keep = ~np.any(np.isnan(y), axis=0)
    return y[:, keep], z[:, keep]


def sample_b(nosamples, bdata):
    """Sample background fluorescence."""
    s2 = np.var(bdata) / bdata.size
    u = np.mean(bdata)
    samples = u + s2 * rng.standard_normal(nosamples)
    if np.any(samples < 0):
        print("Warning: negative background fluorescence.")
    return samples


def get_background_samples(yn, zn, nosamples):
    """Get samples of background fluorescence for GFP and AutoFL."""
    by = sample_b(nosamples, yn)
    bz = sample_b(nosamples, zn)
    return by, bz


def set_up(y, z, ywt, zwt, yn, zn):
    """Define stats_dict."""
    # raguess = np.max((zwt - np.median(zn)) / (ywt - np.median(yn)), axis=1)
    # raguess = raguess[~np.isinf(raguess)]
    stats_dict = {
        "x0": None,
        "gmax": np.max(y),
        "amax": np.max(ywt),
        "ramin": 0,  # np.min(raguess[raguess > 0]),
        "ramax": 1,  # np.max(raguess[raguess > 0]),
        "seq_prior_mu": None,
        "seq_prior_hess": None,
    }
    return stats_dict


def minus_log_prob_array(theta, stats_dict):
    """Get log normal probability."""
    g, ra, a = theta
    rg = stats_dict["rg"]
    n = stats_dict["n"]
    sy = stats_dict["sy"]
    sz = stats_dict["sz"]
    sywt = stats_dict["sywt"]
    szwt = stats_dict["szwt"]
    by = stats_dict["by"]
    bz = stats_dict["bz"]
    bywt = stats_dict["by"]
    bzwt = stats_dict["bz"]
    ly = stats_dict["ly"]
    lz = stats_dict["lz"]
    lywt = stats_dict["lywt"]
    lzwt = stats_dict["lzwt"]
    mlp_v = np.sum(
        n * np.log(sy * sz * sywt * szwt)
        + (np.log(a + g + by) - ly) ** 2 / (2 * sy**2)
        + (np.log(a * ra + g * rg + bz) - lz) ** 2 / (2 * sz**2)
        + (np.log(a + bywt) - lywt) ** 2 / (2 * sywt**2)
        + (np.log(a * ra + bzwt) - lzwt) ** 2 / (2 * szwt**2),
        axis=0,
    )
    return mlp_v


def minus_log_prob(theta, stats_dict):
    """Find joint probability averaged over background fluorescence."""
    mlp_v = minus_log_prob_array(theta, stats_dict)
    mlp = average_background_fluorescence(mlp_v)
    return mlp


def average_background_fluorescence(mlp, deriv=None):
    """Average background fluorescence."""
    norm = np.min(mlp)
    # normalise to prevent underflows, exponentiate, then average
    norm_prob = np.exp(norm - mlp)
    if deriv is None:
        # revert normalisation and return -log(probability)
        minus_log_prob = norm - np.log(np.mean(norm_prob))
        return minus_log_prob
    else:
        # return weighted average of a derivative array
        deriv_average = np.sum(deriv * norm_prob) / np.sum(norm_prob)
        return deriv_average


def jac_arrays(theta, stats_dict):
    """Get Jacobian of log normal probability."""
    g, ra, a = theta
    rg = stats_dict["rg"]
    sy = stats_dict["sy"]
    sz = stats_dict["sz"]
    sywt = stats_dict["sywt"]
    szwt = stats_dict["szwt"]
    by = stats_dict["by"]
    bz = stats_dict["bz"]
    bywt = stats_dict["by"]
    bzwt = stats_dict["bz"]
    ly = stats_dict["ly"]
    lz = stats_dict["lz"]
    lywt = stats_dict["lywt"]
    lzwt = stats_dict["lzwt"]
    jac_v = np.zeros(3, dtype="object")
    y_v = (np.log(a + g + by) - ly) / (sy**2 * (a + g + by))
    z_v = (np.log(a * ra + g * rg + bz) - lz) / (
        sz**2 * (a * ra + g * rg + bz)
    )
    ywt_v = (np.log(a + bywt) - lywt) / (sywt**2 * (a + bywt))
    zwt_v = (np.log(a * ra + bzwt) - lzwt) / (szwt**2 * (a * ra + bzwt))
    # with respect to g
    jac_v[0] = np.sum(y_v + rg * z_v, axis=0)
    # with respect to ra
    jac_v[1] = np.sum(a * (z_v + zwt_v), axis=0)
    # with respect to a
    jac_v[2] = np.sum(y_v + ywt_v + ra * (z_v + zwt_v), axis=0)
    return jac_v


def jac(theta, stats_dict, return_jac_v=False):
    """Get Jacobian averaged over background fluorescence."""
    jac_v = jac_arrays(theta, stats_dict)
    # average background fluorescence
    mlp_v = minus_log_prob_array(theta, stats_dict)
    av_jac = np.array(
        [average_background_fluorescence(mlp_v, jac_v[i]) for i in range(3)]
    )
    if return_jac_v:
        return av_jac, jac_v
    else:
        return av_jac


def hess_arrays(theta, stats_dict):
    """Get Hessian of log normal probability."""
    g, ra, a = theta
    rg = stats_dict["rg"]
    sy = stats_dict["sy"]
    sz = stats_dict["sz"]
    sywt = stats_dict["sywt"]
    szwt = stats_dict["szwt"]
    by = stats_dict["by"]
    bz = stats_dict["bz"]
    bywt = stats_dict["by"]
    bzwt = stats_dict["bz"]
    ly = stats_dict["ly"]
    lz = stats_dict["lz"]
    lywt = stats_dict["lywt"]
    lzwt = stats_dict["lzwt"]
    hess_v = np.zeros((3, 3), dtype="object")
    y_v = (1 - np.log(a + g + by) + ly) / (sy**2 * (a + g + by) ** 2)
    z_v = (1 - np.log(a * ra + g * rg + bz) + lz) / (
        sz**2 * (a * ra + g * rg + bz) ** 2
    )
    ywt_v = (1 - np.log(a + bywt) + lywt) / (sywt**2 * (a + bywt) ** 2)
    zwt_v = (1 - np.log(a * ra + bzwt) + lzwt) / (
        szwt**2 * (a * ra + bzwt) ** 2
    )
    z_vv = (np.log(a * ra + g * rg + bz) - lz) / (
        sz**2 * (a * ra + g * rg + bz)
    )
    zwt_vv = (np.log(a * ra + bzwt) - lzwt) / (szwt**2 * (a * ra + bzwt))
    # g, g
    hess_v[0, 0] = y_v + rg**2 * z_v
    # ra, ra
    hess_v[1, 1] = a**2 * (z_v + zwt_v)
    # a, a
    hess_v[2, 2] = y_v + ywt_v + ra**2 * (z_v + zwt_v)
    # g, ra
    hess_v[0, 1] = a * rg * z_v
    # g, a
    hess_v[0, 2] = y_v + ra * rg * z_v
    # ra, a
    hess_v[1, 2] = a * ra * (z_v + zwt_v) + z_vv + zwt_vv
    # sum over replicates
    for i in range(3):
        for j in range(i, 3):
            hess_v[i, j] = np.sum(hess_v[i, j], axis=0)
            if j > i:
                hess_v[j, i] = hess_v[i, j]
    return hess_v


def hess(theta, stats_dict):
    """Get Hessian averaged over background fluorescence."""
    mlp_v = minus_log_prob_array(theta, stats_dict)
    av_jac, jac_v = jac(theta, stats_dict, return_jac_v=True)
    hess_v = hess_arrays(theta, stats_dict)
    av_hess = np.empty((3, 3))
    for i in range(3):
        for j in range(i, 3):
            av_hess[i, j] = (
                average_background_fluorescence(mlp_v, hess_v[i, j])
                + av_jac[i] * av_jac[j]
                - average_background_fluorescence(mlp_v, jac_v[i] * jac_v[j])
            )
            if j > i:
                av_hess[j, i] = av_hess[i, j]
    return av_hess


def minus_log_prior(theta, stats_dict):
    """Get minus log of the prior probability."""
    g, ra, a = theta
    mlp = 0
    if stats_dict["seq_prior_hess"] is not None:
        mu = stats_dict["seq_prior_mu"]
        h = stats_dict["seq_prior_hess"]
        mlp = (
            (g - mu[0]) ** 2 * h[0, 0]
            + 2 * (g - mu[0]) * (ra - mu[1]) * h[0, 1]
            + 2 * (g - mu[0]) * (a - mu[2]) * h[0, 2]
            + (ra - mu[1]) ** 2 * h[1, 1]
            + 2 * (ra - mu[1]) * (a - mu[2]) * h[1, 2]
            + (a - mu[2]) ** 2 * h[2, 2]
        ) / 2
    return mlp


def jac_prior(theta, stats_dict):
    """Get Jacobian for -log prior."""
    g, ra, a = theta
    jac = np.zeros(3)
    if stats_dict["seq_prior_mu"] is not None:
        mu = stats_dict["seq_prior_mu"]
        h = stats_dict["seq_prior_hess"]
        jac[0] = (
            g * h[0, 0]
            + a * h[0, 2]
            - h[0, 0] * mu[0]
            - h[0, 1] * mu[1]
            - h[0, 2] * mu[2]
            + h[0, 1] * ra
        )
        jac[1] = (
            g * h[0, 1]
            + a * h[1, 2]
            - h[0, 1] * mu[0]
            - h[1, 1] * mu[1]
            - h[1, 2] * mu[2]
            + h[1, 1] * ra
        )
        jac[2] = (
            g * h[0, 2]
            + a * h[2, 2]
            - h[0, 2] * mu[0]
            - h[1, 2] * mu[1]
            - h[2, 2] * mu[2]
            + h[1, 2] * ra
        )
    return jac


def hess_prior(theta, stats_dict):
    """Get Hessian for -log prior."""
    hess = np.zeros((3, 3))
    if stats_dict["seq_prior_hess"] is not None:
        hess = stats_dict["seq_prior_hess"]
    return hess


def set_up_minimization(stats_dict):
    """Set up bounds and initial guess."""
    g_bounds = (0, stats_dict["gmax"])
    ra_bounds = (stats_dict["ramin"], stats_dict["ramax"])
    a_bounds = (0, stats_dict["amax"])
    bounds = [g_bounds, ra_bounds, a_bounds]
    x0_o = np.array(
        [
            np.median(np.exp(stats_dict["ly"])),
            np.median(np.exp(stats_dict["lzwt"] - stats_dict["lywt"])),
            np.median(np.exp(stats_dict["lywt"])),
        ]
    )
    if stats_dict["x0"] is None:
        x0 = x0_o
    else:
        x0 = stats_dict["x0"]
    return bounds, x0, x0_o


def find_mode(stats_dict, no_attempts, prior=False):
    """
    Find most probable value of g and ra.

    Tried with a sequential approach, using the a Gaussian
    approximation of the current posterior as the prior
    for the future one, but performs worse, perhaps because
    of the wide hessian.
    """
    bounds, x0, x0_o = set_up_minimization(stats_dict)
    min_mlp = np.inf
    mode = None
    rands = rng.standard_normal((no_attempts, 3))
    for i in range(no_attempts):
        if i < int(no_attempts / 2):
            # start from proceeding optimum
            sampled_x0 = rands[i, :] * 10 * np.sqrt(x0) + x0
        else:
            # start from a new guess
            sampled_x0 = rands[i, :] * 10 * np.sqrt(x0_o) + x0_o
        sampled_x0[sampled_x0 < 0] = 0.01
        if prior:
            res = minimize(
                lambda x: minus_log_prob(x, stats_dict)
                + minus_log_prior(x, stats_dict),
                x0=sampled_x0,
                bounds=bounds,
                jac=lambda x: jac(x, stats_dict) + jac_prior(x, stats_dict),
                hess=lambda x: hess(x, stats_dict) + hess_prior(x, stats_dict),
                method="L-BFGS-B",
            )
        else:
            res = minimize(
                lambda x: minus_log_prob(x, stats_dict),
                x0=sampled_x0,
                bounds=bounds,
                jac=lambda x: jac(x, stats_dict),
                hess=lambda x: hess(x, stats_dict),
                method="L-BFGS-B",
            )
        if res.success:
            if res.fun < min_mlp:
                mode = res.x
                min_mlp = res.fun
                stats_dict["x0"] = res.x
                hit_boundary = [
                    j
                    for j in range(3)
                    if mode[j] == bounds[j][0] or mode[j] == bounds[j][1]
                ]
                if not hit_boundary:
                    # update prior following sequential bayes
                    stats_dict["seq_prior_mu"] = res.x
                    stats_dict["seq_prior_hess"] = hess(res.x, stats_dict)
    if mode is None:
        print(" Warning: Maximising posterior probability failed.")
        mode = np.nan * np.ones(3)
        stats_dict["x0"] = None
    return mode


def correctauto_bayesian(
    self,
    f,
    refstrain,
    flcvfn,
    bd,
    max_data_pts,
    nosamples_for_bg,
    no_minimisation_attempts,
    nosamples,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
):
    """
    Correct fluorescence for auto- and background fluorescence.

    Use a Bayesian method to correct for autofluorescence from fluorescence
    measurements at two wavelengths and for background fluorescence.

    Implement demixing following Lichten ... Swain, BMC Biophys 2014.

    Integrate over autofluorescence exactly and other nuisance variable
    by sampling.
    """
    print("Using Bayesian approach for two fluorescence wavelengths.")
    print(f"Correcting autofluorescence using {f[0]} and {f[1]}.")
    bname = "bc" + f[0]
    bd_default = {0: (-2, 8), 1: (-2, 4), 2: (-4, 6)}
    if bd is not None:
        bdn = gu.mergedicts(original=bd_default, update=bd)
    else:
        bdn = bd_default
    for e in sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    ):
        for c in sunder.getset(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            labeltype="condition",
            nonull=True,
            nomedia=True,
        ):
            # get data for reference strain
            # y for emission at 525; z for emission at 585
            _, (ywt, zwt) = sunder.extractwells(
                self.r, self.s, e, c, refstrain, f
            )
            ywt, zwt = de_nan(ywt, zwt)
            # get data for Null
            _, (yn, zn) = sunder.extractwells(self.r, self.s, e, c, "Null", f)
            yn, zn = de_nan(yn, zn)
            # check sufficient replicates
            if (ywt.shape[1] < 3) or (yn.shape[1] < 3):
                raise Exception(
                    f"There are less than three replicates for the {refstrain}"
                    " or Null strains."
                )
            for s in sunder.getset(
                self,
                strains,
                strainincludes,
                strainexcludes,
                labeltype="strain",
                nonull=True,
            ):
                if (
                    s != refstrain
                    and f"{s} in {c}" in self.allstrainsconditions[e]
                ):
                    # get data for tagged strain
                    t, (y, z, od) = sunder.extractwells(
                        self.r, self.s, e, c, s, f.copy() + ["OD"]
                    )
                    y, z = de_nan(y, z)
                    if y.size == 0 or z.size == 0:
                        print(f"Warning: No data found for {e}: {s} in {c}!!")
                        continue
                    if y.shape[1] < 3:
                        raise Exception(
                            "There are less than three replicates"
                            f" for {e}: {s} in {c}."
                        )
                    print(f"{e}: {s} in {c}")
                    # correct autofluorescence for each time point
                    predicted_fl = np.zeros(t.size)
                    stats_dict = set_up(y, z, ywt, zwt, yn, zn)
                    stats_dict["rg"] = self._gamma
                    stats_dict["n"] = y.shape[1]
                    for i in tqdm(range(t.size)):
                        stats_dict["by"], stats_dict["bz"] = (
                            get_background_samples(
                                yn[i, :], zn[i, :], nosamples_for_bg
                            )
                        )
                        stats_dict["sy"] = np.std(np.log(y[i, :]))
                        stats_dict["sz"] = np.std(np.log(z[i, :]))
                        stats_dict["sywt"] = np.std(np.log(ywt[i, :]))
                        stats_dict["szwt"] = np.std(np.log(zwt[i, :]))
                        stats_dict["ly"] = np.log(y[i, :])[:, None]
                        stats_dict["lz"] = np.log(z[i, :])[:, None]
                        stats_dict["lywt"] = np.log(ywt[i, :])[:, None]
                        stats_dict["lzwt"] = np.log(zwt[i, :])[:, None]
                        posterior_mode = find_mode(
                            stats_dict, no_minimisation_attempts
                        )
                        predicted_fl[i] = posterior_mode[0]
                    # smooth with GP and add to data frames
                    print("Smoothing...")
                    flgp, _ = runfitderiv(
                        self,
                        t,
                        predicted_fl,
                        f"{bname}",
                        f"d/dt_{bname}",
                        experiment=e,
                        condition=c,
                        strain=s,
                        bd=bdn,
                        cvfn=flcvfn,
                        logs=False,
                        figs=True,
                        plotlocalmax=False,
                        max_data_pts=max_data_pts,
                    )
                    # sample ODs using GPs
                    lod_samples = omcorr.sample_ODs_with_GP(
                        self, e, c, s, t, od, nosamples
                    )
                    od_samples = np.exp(lod_samples)
                    # find samples of fluorescence per OD
                    smooth_fl_samples = flgp.fitderivsample(nosamples)[0]
                    flperod = smooth_fl_samples / od_samples
                    # store results for fluorescence per OD
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": t,
                        f"{bname}perOD": np.mean(flperod, 1),
                        f"{bname}perOD_err": omcorr.nanstdzeros2nan(
                            flperod, 1
                        ),
                    }
                    # add to data frames
                    omcorr.addtodataframes(self, autofdict, bname)



# A Bayesian version of the method in Lichten ... Swain, BMC Biophys 2014
import numpy as np
from scipy.optimize import minimize

import omniplate.corrections as omcorr
import omniplate.omgenutils as gu
import omniplate.sunder as sunder
from omniplate.runfitderiv import runfitderiv

rng = np.random.default_rng()

# notation follows Lichten ... Swain
# GFP is denoted y; AutoFL is denoted z.
# The reference strain is denoted WT.


def de_nan(y, z):
    """Remove any replicates with NaN."""
    # NaNs are generated because experiments have different durations
    keep = ~np.any(np.isnan(y), axis=0)
    return y[:, keep], z[:, keep]


def sample_b(nosamples, bdata):
    """Sample background fluorescence."""
    s2 = np.var(bdata) / bdata.size
    u = np.mean(bdata)
    samples = u + s2 * rng.standard_normal(nosamples)
    if np.any(samples < 0):
        print("Warning: negative background fluorescence.")
    return samples


def get_background_samples(yn, zn, nosamples):
    """Get samples of background fluorescence for GFP and AutoFL."""
    by = sample_b(nosamples, yn)
    bz = sample_b(nosamples, zn)
    return by, bz


def update_stats_dict(sd, y, z, ywt, zwt, yn, zn, rg, by, bz):
    """Update with statistics needed for sampling."""
    sd["rg"] = rg
    sd["by"] = by
    sd["bz"] = bz
    # for wild type strain
    sd["Sy_wt"], sd["Ty_wt"], sd["Wy_wt"] = get_sufficient_stats(ywt)
    sd["Sz_wt"], sd["Tz_wt"], sd["Wz_wt"] = get_sufficient_stats(zwt)
    sd["sy_wt"] = np.std(ywt)
    sd["sz_wt"] = np.std(zwt)
    sd["n_wt"] = ywt.size
    sd["rguess"] = ra_guess(ywt, zwt)
    # for tagged strain
    sd["Sy"], sd["Ty"], sd["Wy"] = get_sufficient_stats(y)
    sd["Sz"], sd["Tz"], sd["Wz"] = get_sufficient_stats(z)
    sd["sy"] = np.std(y)
    sd["sz"] = np.std(z)
    sd["n"] = y.size
    sd["gguess"] = g_guess(y, z, ywt, zwt, rg)
    return sd


def get_sufficient_stats(y):
    """Get sufficient statistics generated by integrating autofluorescence."""
    S = np.sum(y**2)
    T = np.sum(y)
    W = np.sum(
        [y[i] * y[j] for i in range(y.size) for j in range(i + 1, y.size)]
    )
    return S, T, W


def ra_guess(ywt, zwt):
    """Get mean-field estimate of ra."""
    return np.mean(zwt / ywt)


def g_guess(y, z, ywt, zwt, rg):
    """Get mean-field estimate of GFP."""
    ra = ra_guess(ywt, zwt)
    gs = (ra * y - z) / (ra - rg)
    g_guess = np.nanmax([0.01, np.mean(gs[gs > 0])])
    return g_guess


def find_alpha(r, sy, sz, n, ty, tz, Sy, Ty, Wy, Sz, Tz, Wz):
    """Find exponent after integrating out autofluorescence."""
    r = np.asarray(r)
    if r.size > 1:
        r = r[:, None]
    alpha = (
        sz**4 * ((n - 1) * Sy - 2 * Wy)
        + r**2 * sy**4 * ((n - 1) * Sz - 2 * Wz)
        + sy**2
        * sz**2
        * (
            n * r**2 * (n * ty**2 + Sy - 2 * ty * Ty)
            + n * (n * tz**2 + Sz - 2 * tz * Tz)
            - 2 * r * (n * ty - Ty) * (n * tz - Tz)
        )
    )
    return alpha


def mlp(alpha, r, sy, sz, n):
    """Find minus log of probability after integrating autofluorescence."""
    r = np.asarray(r)
    if r.size > 1:
        r = r[:, None]
    mlp = (
        alpha / (2 * n * sy**2 * sz**2 * (r**2 * sy**2 + sz**2))
        + (n - 1) * np.log(sy)
        + (n - 1) * np.log(sz)
        + np.log(r**2 * sy**2 + sz**2) / 2
    )
    return mlp


def minus_log_int_prob_ra(ra, stats_dict):
    """Estimate interior probability of ra."""
    ra = np.asarray(ra)
    alpha = find_alpha(
        ra,
        stats_dict["sy_wt"],
        stats_dict["sz_wt"],
        stats_dict["n_wt"],
        stats_dict["by"],
        stats_dict["bz"],
        stats_dict["Sy_wt"],
        stats_dict["Ty_wt"],
        stats_dict["Wy_wt"],
        stats_dict["Sz_wt"],
        stats_dict["Tz_wt"],
        stats_dict["Wz_wt"],
    )
    minus_log_prob = mlp(
        alpha,
        ra,
        stats_dict["sy_wt"],
        stats_dict["sz_wt"],
        stats_dict["n_wt"],
    )
    return minus_log_prob


def minus_log_int_prob_g_ra(g, ra, stats_dict):
    """Estimate interior probability for a scalar g and ra."""
    alpha = find_alpha(
        ra,
        stats_dict["sy"],
        stats_dict["sz"],
        stats_dict["n"],
        g + stats_dict["by"],
        stats_dict["rg"] * g + stats_dict["bz"],
        stats_dict["Sy"],
        stats_dict["Ty"],
        stats_dict["Wy"],
        stats_dict["Sz"],
        stats_dict["Tz"],
        stats_dict["Wz"],
    )
    minus_log_prob = mlp(
        alpha,
        ra,
        stats_dict["sy"],
        stats_dict["sz"],
        stats_dict["n"],
    )
    return minus_log_prob


def minus_log_prob_g_ra_array(theta, stats_dict):
    """Find joint probability of a scalar g and ra for each by and bz."""
    g, ra = theta
    mlp = minus_log_int_prob_g_ra(g, ra, stats_dict) + minus_log_int_prob_ra(
        ra, stats_dict
    )
    return mlp


def minus_log_prob(theta, stats_dict):
    """Find joint probability of a scalar g and ra averaged over background."""
    mlp = minus_log_prob_g_ra_array(theta, stats_dict)
    minus_log_prob = average_background_fluorescence(mlp)
    return minus_log_prob


def average_background_fluorescence(mlp, deriv=None):
    """Average background fluorescence."""
    norm = np.min(mlp)
    # normalise to prevent underflows, exponentiate, then average
    norm_prob = np.exp(norm - mlp)
    if deriv is None:
        # revert normalisation and return -log(probability)
        minus_log_prob = norm - np.log(np.mean(norm_prob))
        return minus_log_prob
    else:
        # return weighted average of a derivative array
        deriv_average = np.sum(deriv * norm_prob) / np.sum(norm_prob)
        return deriv_average


def find_mode(stats_dict, bnd_factor=100):
    """Find most probable value of g and ra."""
    rmin = stats_dict["rguess"] / bnd_factor
    rmax = stats_dict["rguess"] * bnd_factor
    gmin = stats_dict["gguess"] / bnd_factor
    gmax = stats_dict["gguess"] * bnd_factor
    # aside
    # res = minimize(
    #     lambda x: average_background_fluorescence(
    #         minus_log_int_prob_ra(x, stats_dict)
    #     ),
    #     x0=stats_dict["rguess"],
    #     bounds=((rmin, rmax),),
    #     method="L-BFGS-B",
    # )
    # breakpoint()
    bounds = ((gmin, gmax), (rmin, rmax))
    if stats_dict["seq_prior_mu"] is None:
        x0 = [stats_dict["gguess"], stats_dict["rguess"]]
    else:
        x0 = stats_dict["seq_prior_mu"]
    min_mlp = np.inf
    mode = None
    no_attempts = 10
    rands = rng.standard_normal((no_attempts, 2))
    for i in range(no_attempts):
        x0 = rands[i, :] * np.sqrt(x0) + x0
        res = minimize(
            lambda x: minus_log_prob(x, stats_dict)
            + minus_log_prior(x, stats_dict),
            x0=x0,
            bounds=bounds,
            jac=lambda x: jac(x, stats_dict) + jac_prior(x, stats_dict),
            hess=lambda x: hess(x, stats_dict) + hess_prior(x, stats_dict),
            method="L-BFGS-B",
        )
        if res.success:
            if res.fun < min_mlp:
                mode = res.x
                for j in range(2):
                    if mode[j] == bounds[j][0] or mode[j] == bounds[j][1]:
                        print(f"parameter {j} at bound {mode[j]}")
                stats_dict["seq_prior_mu"] = mode
                stats_dict["seq_prior_hess"] = hess(
                    mode, stats_dict
                ) + hess_prior(mode, stats_dict)
                min_mlp = res.fun
    if mode is None:
        breakpoint()
    return mode


def correctauto_bayesian(
    self,
    f,
    refstrain,
    flcvfn,
    bd,
    max_data_pts,
    nosamples_for_fl,
    nosamples_for_ra,
    nosamples_for_bg,
    nosamples,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
):
    """
    Correct fluorescence for auto- and background fluorescence.

    Use a Bayesian method to correct for autofluorescence from fluorescence
    measurements at two wavelengths and for background fluorescence.

    Implement demixing following Lichten ... Swain, BMC Biophys 2014.

    Integrate over autofluorescence exactly and other nuisance variable
    by sampling.
    """
    print("Using Bayesian approach for two fluorescence wavelengths.")
    print(f"Correcting autofluorescence using {f[0]} and {f[1]}.")
    rg = self._gamma
    bname = "bc" + f[0]
    bd_default = {0: (2, 8), 1: (-2, 4), 2: (5, 8)}
    if bd is not None:
        bdn = gu.mergedicts(original=bd_default, update=bd)
    else:
        bdn = bd_default
    for e in sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        "experiment",
        nonull=True,
    ):
        for c in sunder.getset(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            labeltype="condition",
            nonull=True,
            nomedia=True,
        ):
            # get data for reference strain
            # y for emission at 525; z for emission at 585
            _, (ywt, zwt) = sunder.extractwells(
                self.r, self.s, e, c, refstrain, f
            )
            ywt, zwt = de_nan(ywt, zwt)
            # get data for Null
            _, (yn, zn) = sunder.extractwells(self.r, self.s, e, c, "Null", f)
            yn, zn = de_nan(yn, zn)
            # check sufficient replicates
            if (ywt.shape[1] < 3) or (yn.shape[1] < 3):
                raise Exception(
                    f"There are less than three replicates for the {refstrain}"
                    " or Null strains."
                )
            for s in sunder.getset(
                self,
                strains,
                strainincludes,
                strainexcludes,
                labeltype="strain",
                nonull=True,
            ):
                if (
                    s != refstrain
                    and f"{s} in {c}" in self.allstrainsconditions[e]
                ):
                    # get data for tagged strain
                    t, (y, z, od) = sunder.extractwells(
                        self.r, self.s, e, c, s, f.copy() + ["OD"]
                    )
                    y, z = de_nan(y, z)
                    if y.size == 0 or z.size == 0:
                        print(f"Warning: No data found for {e}: {s} in {c}!!")
                        continue
                    if y.shape[1] < 3:
                        raise Exception(
                            "There are less than three replicates"
                            f" for {e}: {s} in {c}."
                        )
                    print(f" {e}: {s} in {c}")
                    # correct autofluorescence for each time point
                    predicted_fl = np.zeros(t.size)
                    # for sequential priors
                    stats_dict = {"seq_prior_mu": None, "seq_prior_hess": None}
                    for i in range(t.size):
                        by, bz = get_background_samples(
                            yn[i, :], zn[i, :], nosamples_for_bg
                        )
                        if i == 10:
                            breakpoint()
                        stats_dict = update_stats_dict(
                            stats_dict,
                            y[i, :],
                            z[i, :],
                            ywt[i, :],
                            zwt[i, :],
                            yn[i, :],
                            zn[i, :],
                            rg,
                            by,
                            bz,
                        )
                        predicted_fl[i] = find_mode(stats_dict)[0]
                    # smooth with GP and add to data frames
                    if False:
                        print("Smoothing...")
                        flgp, _ = runfitderiv(
                            self,
                            t,
                            predicted_fl,
                            f"{bname}",
                            f"d/dt_{bname}",
                            e,
                            c,
                            s,
                            bd=bdn,
                            cvfn=flcvfn,
                            logs=True,
                            figs=True,
                            max_data_pts=max_data_pts,
                        )
                    # sample ODs using GPs
                    lod_samples = omcorr.sample_ODs_with_GP(
                        self, e, c, s, t, od, nosamples
                    )
                    od_samples = np.exp(lod_samples)
                    # find samples of fluorescence per OD
                    smooth_fl_samples = predicted_fl[:, None]
                    # smooth_fl_samples = flgp.fitderivsample(nosamples)[0]
                    flperod = smooth_fl_samples / od_samples
                    # store results for fluorescence per OD
                    autofdict = {
                        "experiment": e,
                        "condition": c,
                        "strain": s,
                        "time": t,
                        f"{bname}perOD": np.nanmean(flperod, 1),
                        f"{bname}perOD_err": omcorr.nanstdzeros2nan(
                            flperod, 1
                        ),
                    }
                    # add to data frames
                    omcorr.addtodataframes(self, autofdict, bname)


def jac_arrays(theta, stats_dict):
    """Find Jacobian of minus_log_total_prob for each by and bz."""
    g, ra = theta
    rg = stats_dict["rg"]
    by = stats_dict["by"]
    bz = stats_dict["bz"]
    Sywt = stats_dict["Sy_wt"]
    Tywt = stats_dict["Ty_wt"]
    Wywt = stats_dict["Wy_wt"]
    Szwt = stats_dict["Sz_wt"]
    Tzwt = stats_dict["Tz_wt"]
    Wzwt = stats_dict["Wz_wt"]
    sywt = stats_dict["sy_wt"]
    szwt = stats_dict["sz_wt"]
    Sy = stats_dict["Sy"]
    Ty = stats_dict["Ty"]
    Wy = stats_dict["Wy"]
    Sz = stats_dict["Sz"]
    Tz = stats_dict["Tz"]
    Wz = stats_dict["Wz"]
    sy = stats_dict["sy"]
    sz = stats_dict["sz"]
    nwt = stats_dict["n_wt"]
    n = stats_dict["n"]
    # define jacobian over background fluorescence
    jac_v = np.empty(2, dtype="object")
    jac_v[0] = (
        (ra - rg)
        * (-(bz * n) + by * n * ra + g * n * ra - g * n * rg - ra * Ty + Tz)
    ) / (ra**2 * sy**2 + sz**2)
    jac_v[1] = (
        (ra * sy**2) / (ra**2 * sy**2 + sz**2)
        + (ra * sywt**2) / (ra**2 * sywt**2 + szwt**2)
        + (
            sy**2
            * sz**2
            * (
                2 * n * ra * ((by + g) ** 2 * n + Sy - 2 * (by + g) * Ty)
                - 2 * ((by + g) * n - Ty) * (bz * n + g * n * rg - Tz)
            )
            + 2 * ra * sy**4 * ((-1 + n) * Sz - 2 * Wz)
        )
        / (2.0 * n * sy**2 * sz**2 * (ra**2 * sy**2 + sz**2))
        - (
            ra
            * (
                sy**2
                * sz**2
                * (
                    n * ra**2 * ((by + g) ** 2 * n + Sy - 2 * (by + g) * Ty)
                    - 2 * ra * ((by + g) * n - Ty) * (bz * n + g * n * rg - Tz)
                    + n
                    * (n * (bz + g * rg) ** 2 + Sz - 2 * (bz + g * rg) * Tz)
                )
                + sz**4 * ((-1 + n) * Sy - 2 * Wy)
                + ra**2 * sy**4 * ((-1 + n) * Sz - 2 * Wz)
            )
        )
        / (n * (ra**2 * sy**2 * sz + sz**3) ** 2)
        + (
            sywt**2
            * szwt**2
            * (
                2 * nwt * ra * (by**2 * nwt + Sywt - 2 * by * Tywt)
                - 2 * (by * nwt - Tywt) * (bz * nwt - Tzwt)
            )
            + 2 * ra * sywt**4 * ((-1 + nwt) * Szwt - 2 * Wzwt)
        )
        / (2.0 * nwt * sywt**2 * szwt**2 * (ra**2 * sywt**2 + szwt**2))
        - (
            ra
            * (
                sywt**2
                * szwt**2
                * (
                    nwt * ra**2 * (by**2 * nwt + Sywt - 2 * by * Tywt)
                    - 2 * ra * (by * nwt - Tywt) * (bz * nwt - Tzwt)
                    + nwt * (bz**2 * nwt + Szwt - 2 * bz * Tzwt)
                )
                + szwt**4 * ((-1 + nwt) * Sywt - 2 * Wywt)
                + ra**2 * sywt**4 * ((-1 + nwt) * Szwt - 2 * Wzwt)
            )
        )
        / (nwt * (ra**2 * sywt**2 * szwt + szwt**3) ** 2)
    )
    return jac_v


def jac(theta, stats_dict, return_jac_v=False):
    """Get Jacobian averaged over background fluorescence."""
    jac_v = jac_arrays(theta, stats_dict)
    # average background fluorescence
    mlp = minus_log_prob_g_ra_array(theta, stats_dict)
    av_jac = np.empty(2)
    av_jac[0] = average_background_fluorescence(mlp, jac_v[0])
    av_jac[1] = average_background_fluorescence(mlp, jac_v[1])
    if return_jac_v:
        return av_jac, jac_v
    else:
        return av_jac


def hess_arrays(theta, stats_dict):
    """Find Hessian of minus_log_total_prob for each by and bz."""
    g, ra = theta
    rg = stats_dict["rg"]
    by = stats_dict["by"]
    bz = stats_dict["bz"]
    Sywt = stats_dict["Sy_wt"]
    Tywt = stats_dict["Ty_wt"]
    Wywt = stats_dict["Wy_wt"]
    Szwt = stats_dict["Sz_wt"]
    Tzwt = stats_dict["Tz_wt"]
    Wzwt = stats_dict["Wz_wt"]
    sywt = stats_dict["sy_wt"]
    szwt = stats_dict["sz_wt"]
    Sy = stats_dict["Sy"]
    Ty = stats_dict["Ty"]
    Wy = stats_dict["Wy"]
    Sz = stats_dict["Sz"]
    Tz = stats_dict["Tz"]
    Wz = stats_dict["Wz"]
    sy = stats_dict["sy"]
    sz = stats_dict["sz"]
    nwt = stats_dict["n_wt"]
    n = stats_dict["n"]
    # define hessian averaging over background fluorescence
    hess_v = np.empty((2, 2), dtype="object")
    hess_v[0, 0] = (n * (ra - rg) ** 2) / (ra**2 * sy**2 + sz**2)
    hess_v[0, 1] = (
        2 * g * n * ra**2 * rg * sy**2
        - 2 * g * n * ra * rg**2 * sy**2
        + 2 * g * n * ra * sz**2
        - 2 * g * n * rg * sz**2
        + bz * n * (ra**2 * sy**2 - 2 * ra * rg * sy**2 - sz**2)
        + by * n * (ra**2 * rg * sy**2 + 2 * ra * sz**2 - rg * sz**2)
        - ra**2 * rg * sy**2 * Ty
        - 2 * ra * sz**2 * Ty
        + rg * sz**2 * Ty
        - ra**2 * sy**2 * Tz
        + 2 * ra * rg * sy**2 * Tz
        + sz**2 * Tz
    ) / (ra**2 * sy**2 + sz**2) ** 2
    hess_v[1, 0] = hess_v[0, 1]
    hess_v[1, 1] = (
        (-2 * ra**2 * sy**4) / (ra**2 * sy**2 + sz**2) ** 2
        + sy**2 / (ra**2 * sy**2 + sz**2)
        - (2 * ra**2 * sywt**4) / (ra**2 * sywt**2 + szwt**2) ** 2
        + sywt**2 / (ra**2 * sywt**2 + szwt**2)
        + (
            (ra**2 * sy**2 + sz**2) ** 2
            * (
                2
                * n
                * sy**2
                * sz**2
                * ((by + g) ** 2 * n + Sy - 2 * (by + g) * Ty)
                + 2 * sy**4 * ((-1 + n) * Sz - 2 * Wz)
            )
            - 4
            * ra
            * sy**2
            * (ra**2 * sy**2 + sz**2)
            * (
                sy**2
                * sz**2
                * (
                    2 * n * ra * ((by + g) ** 2 * n + Sy - 2 * (by + g) * Ty)
                    - 2 * ((by + g) * n - Ty) * (bz * n + g * n * rg - Tz)
                )
                + 2 * ra * sy**4 * ((-1 + n) * Sz - 2 * Wz)
            )
            + 2
            * (3 * ra**2 * sy**4 - sy**2 * sz**2)
            * (
                sy**2
                * sz**2
                * (
                    n * ra**2 * ((by + g) ** 2 * n + Sy - 2 * (by + g) * Ty)
                    - 2 * ra * ((by + g) * n - Ty) * (bz * n + g * n * rg - Tz)
                    + n
                    * (n * (bz + g * rg) ** 2 + Sz - 2 * (bz + g * rg) * Tz)
                )
                + sz**4 * ((-1 + n) * Sy - 2 * Wy)
                + ra**2 * sy**4 * ((-1 + n) * Sz - 2 * Wz)
            )
        )
        / (2.0 * n * sy**2 * sz**2 * (ra**2 * sy**2 + sz**2) ** 3)
        + (
            -3 * ra**2 * sywt**2 * Sywt * szwt**2
            + Sywt * szwt**4
            + bz**2 * nwt**2 * sywt**2 * (3 * ra**2 * sywt**2 - szwt**2)
            + by**2 * nwt**2 * szwt**2 * (-3 * ra**2 * sywt**2 + szwt**2)
            + 3 * ra**2 * sywt**4 * Szwt
            - sywt**2 * szwt**2 * Szwt
            - 2 * ra**3 * sywt**4 * Tywt * Tzwt
            + 6 * ra * sywt**2 * szwt**2 * Tywt * Tzwt
            + 2
            * bz
            * nwt
            * sywt**2
            * (
                -(by * nwt * ra**3 * sywt**2)
                + 3 * by * nwt * ra * szwt**2
                + ra**3 * sywt**2 * Tywt
                - 3 * ra * szwt**2 * Tywt
                - 3 * ra**2 * sywt**2 * Tzwt
                + szwt**2 * Tzwt
            )
            + 2
            * by
            * nwt
            * (
                3 * ra**2 * sywt**2 * szwt**2 * Tywt
                - szwt**4 * Tywt
                + ra**3 * sywt**4 * Tzwt
                - 3 * ra * sywt**2 * szwt**2 * Tzwt
            )
            - 6 * ra**2 * sywt**2 * szwt**2 * Wywt
            + 2 * szwt**4 * Wywt
            + 6 * ra**2 * sywt**4 * Wzwt
            - 2 * sywt**2 * szwt**2 * Wzwt
        )
        / (nwt * (ra**2 * sywt**2 + szwt**2) ** 3)
    )
    return hess_v


def hess(theta, stats_dict):
    """Get Hessian averaged over background fluorescence."""
    mlp = minus_log_prob_g_ra_array(theta, stats_dict)
    av_jac, jac_v = jac(theta, stats_dict, return_jac_v=True)
    hess_v = hess_arrays(theta, stats_dict)
    av_hess = np.empty((2, 2))
    for i in range(2):
        for j in range(i, 2):
            av_hess[i, j] = (
                average_background_fluorescence(mlp, hess_v[i, j])
                - average_background_fluorescence(mlp, jac_v[0] * jac_v[1])
                - av_jac[0] * av_jac[1]
            )
    av_hess[1, 0] = av_hess[0, 1]
    return av_hess


def minus_log_prior(theta, stats_dict):
    """Get minus log of the prior probability."""
    g, ra = theta
    # mlp = np.log(g) + np.log(ra)
    mlp = 0
    if stats_dict["seq_prior_hess"] is not None:
        mu = stats_dict["seq_prior_mu"]
        h = stats_dict["seq_prior_hess"]
        mlp += (
            (g - mu[0]) ** 2 * h[0, 0]
            + 2 * (g - mu[0]) * (ra - mu[1]) * h[0, 1]
            + (ra - mu[1]) ** 2 * h[1, 1]
        ) / 2
    return mlp


def jac_prior(theta, stats_dict):
    """Get Jacobian for -log prior."""
    g, ra = theta
    jac = np.empty(2)
    jac[0] = 0  # 1 / g
    jac[1] = 0  # 1 / ra
    if stats_dict["seq_prior_mu"] is not None:
        mu = stats_dict["seq_prior_mu"]
        h = stats_dict["seq_prior_hess"]
        jac[0] += g * h[0, 0] - h[0, 0] * mu[0] + h[0, 1] * (-mu[1] + ra)
        jac[1] += (
            g * h[0, 1] - h[0, 1] * mu[0] - h[1, 1] * mu[1] + h[1, 1] * ra
        )
    return jac


def hess_prior(theta, stats_dict):
    """Get Hessian for -log prior."""
    g, ra = theta
    hess = np.empty((2, 2))
    hess[0, 0] = 0  # -1 / g**2
    hess[1, 1] = 0  # -1 / ra**2
    hess[0, 1] = 0
    if stats_dict["seq_prior_hess"] is not None:
        h = stats_dict["seq_prior_hess"]
        hess[0, 0] += h[0, 0]
        hess[1, 1] += h[1, 1]
        hess[0, 1] += h[0, 1]
    hess[1, 0] = hess[0, 1]
    return hess



from pathlib import Path

import pandas as pd

import omniplate.omerrors as errors
from omniplate.parseSunrise import parseSunrise
from omniplate.parseTecan import parseTecan


def parseplate(
    platereaderfile, platereadertype, wdirpath=".", sheetnumber=0, export=False
):
    """
    Parse plate-reader output into a long dataframe.

    Note that the plate-reader file is assumed to be an Excel file
    and that time is converted into hours.

    Parameters
    ----------
    platereaderfile: str
        The name of the data file created by a platereader.
    platereadertype: str
        The type of plate reader, currently only "Tecan"
        for a Tecan M200 or Tecan F200.
    wdirpath: str, optional
        The path to the platereader_file.
    sheetnumber: integer, optional
        The sheet to read from an excel file.
    export: boolean, optional
        If True, write parsed data to tsv file.

    Example
    -------
    >>> from om_code.parseplate import parseplate
    >>> rdf= parseplate("ExampleData.xlsx", "Tecan", wdirpath="data")
    >>> print(rdf)

                time well      OD   GFP  AutoFL  mCherry
    0       0.000000   A1  0.2555  46.0    18.0     19.0
    1       0.232306   A1  0.2725  45.0    17.0     17.0

    """
    if isinstance(wdirpath, str):
        wdirpath = Path(wdirpath)
    if platereadertype == "tidy":
        print(
            "Columns must be labelled 'time', 'well', 'OD', etc., "
            "and time must be in units of hours."
        )
        try:
            if ".tsv" in platereaderfile:
                rdf = pd.read_csv(
                    str(wdirpath / platereaderfile), sep="\t", index_col=0
                )
            else:
                rdf = pd.read_csv(str(wdirpath / platereaderfile), index_col=0)
        except FileNotFoundError:
            raise errors.FileNotFound(str(wdirpath / platereaderfile))
        if rdf.time.max() > 100:
            print("Warning: time does not appear to be in hours.")
        return rdf
    else:
        # load data
        try:
            dfd = pd.read_excel(str(wdirpath / platereaderfile), sheet_name=sheetnumber)
        except FileNotFoundError:
            raise errors.FileNotFound(str(wdirpath / platereaderfile))
        # create a dict to store data
        rdict = {"time": [], "well": []}
        # parse loaded data frame
        if platereadertype == "Tecan":
            rdict = parseTecan(dfd, rdict)
        elif platereadertype == "Sunrise":
            rdict = parseSunrise(dfd, rdict)
        else:
            raise ValueError(f"{platereadertype} not recognised.")
        # convert parsed dict to data frame
        rdf = pd.DataFrame.from_dict(rdict)
        if export:
            outname = platereaderfile.split(".")[0] + ".tsv"
            print(f"exporting {outname}")
            rdf.to_csv(str(wdirpath / outname), sep="\t")
        return rdf



import numpy as np

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omplot as omplot
import omniplate.sunder as sunder
from omniplate.runfitderiv import runfitderiv


@clogger.log
def getstats(
    self,
    dtype="OD",
    bd=False,
    cvfn="matern",
    empirical_errors=False,
    noruns=10,
    exitearly=True,
    noinits=100,
    nosamples=100,
    logs=True,
    figs=True,
    findareas=False,
    plotlocalmax=True,
    showpeakproperties=False,
    max_data_pts=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
    **kwargs,
):
    """
    Smooth data, find its derivatives, and calculate summary statistics.

    The first and second time derivatives are found, typically of OD,
    using a Gaussian process (Swain et al., 2016).

    The derivatives are stored in the .s dataframe;
    summary statistics are stored in the .sc dataframe.

    Parameters
    ----------
    dtype: string, optional
        The type of data - 'OD', 'GFP', 'cGFPperOD', or 'cGFP' - for
        which the derivatives are to be found. The data must exist in the
        .r or .s dataframes.
    bd: dictionary, optional
        The bounds on the hyperparameters for the Gaussian process.
        For example, bd= {1: [-2,0])} fixes the bounds on the
        hyperparameter controlling flexibility to be 1e-2 and 1e0.
        The default for a Matern covariance function
        is {0: (-5,5), 1: (-4,4), 2: (-5,2)},
        where the first element controls amplitude, the second controls
        flexibility, and the third determines the magnitude of the
        measurement error.
    cvfn: string, optional
        The covariance function used in the Gaussian process, either
        'matern' or 'sqexp' or 'nn'.
    empirical_errors: boolean, optional
        If True, measurement errors are empirically estimated from the
        variance across replicates at each time point and so vary with
        time.
        If False, the magnitude of the measurement error is fit from the
        data assuming that this magnitude is the same at all time points.
    noruns: integer, optional
        The number of attempts made for each fit. Each attempt is made
        with random initial estimates of the hyperparameters within their
        bounds.
    exitearly: boolean, optional
        If True, stop at the first successful fit.
        If False, use the best fit from all successful fits.
    noinits: integer, optional
        The number of random attempts to find a good initial condition
        before running the optimization.
    nosamples: integer, optional
        The number of samples used to calculate errors in statistics by
        bootstrapping.
    logs: boolean, optional
        If True, find the derivative of the log of the data and should be
        True to determine the specific growth rate when dtype= 'OD'.
    figs: boolean, optional
        If True, plot both the fits and inferred derivative.
    findareas: boolean, optional
        If True, find the area under the plot of gr vs OD and the area
        under the plot of OD vs time. Setting to True can make getstats
        slow.
    plotlocalmax: boolean, optional
        If True, mark the highest local maxima found, which is used to
        calculate statistics, on any plots.
    showpeakproperties: boolean, optional
        If True, show properties of any local peaks that have found by
        scipy's find_peaks. Additional properties can be specified as
        kwargs and are passed to find_peaks.
    max_data_pts: integer, optional
        If set, sufficiently large data sets with multiple replicates will
        be subsampled at each time point, randomly picking a smaller
        number of replicates, to reduce the number of data points and so
        run times.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.
    kwargs: for scipy's find_peaks
        To set the minimum property of a peak. e.g. prominence= 0.1 and
        width= 15 (specified in numbers of x-points or y-points and not
        real units).
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html

    Examples
    --------
    >>> p.getstats()
    >>> p.getstats(conditionincludes= 'Gal')
    >>> p.getstats(noruns= 10, exitearly= False)

    If the fits are poor, often changing the bounds on the hyperparameter
    for the measurement error helps:

    >>> p.getstats(bd= {2: (-3,0)})

    References
    ----------
    PS Swain, K Stevenson, A Leary, LF Montano-Gutierrez, IB Clark,
    J Vogel, T Pilizota. (2016). Inferring time derivatives including cell
    growth rates using Gaussian processes. Nat Commun, 7, 1-8.
    """
    admin.check_kwargs(kwargs)
    linalgmax = 5
    warnings = ""
    # variable to be fit
    if logs:
        fitvar = f"log_{dtype}"
    else:
        fitvar = dtype
    # name of derivative of fit variable
    if fitvar == "log_OD":
        derivname = "gr"
    else:
        derivname = f"d/dt_{fitvar}"
    # extract data
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=True,
    )
    # find growth rate and stats
    for e in exps:
        for c in cons:
            for s in strs:
                esc_name = f"{e}: {s} in {c}"
                if f"{s} in {c}" in self.allstrainsconditions[e]:
                    if dtype in self.r.columns:
                        # raw data
                        t, d = sunder.extractwells(
                            self.r, self.s, e, c, s, dtype
                        )
                    elif dtype in self.s.columns:
                        # processed data
                        df = self.s.query(
                            "experiment == @e and condition == @c and "
                            "strain == @s"
                        )
                        # add columns plus and minus err
                        df = omplot.augmentdf(df, dtype)[
                            [dtype, "augtype", "time"]
                        ]
                        piv_df = df.pivot(
                            index="time", columns="augtype", values=dtype
                        )
                        # convert to array for fitderiv
                        d = piv_df.values
                        t = piv_df.index.to_numpy()
                        numberofnans = np.count_nonzero(np.isnan(d))
                        if np.any(numberofnans):
                            print(f"\nWarning: {numberofnans} NaNs in data")
                    else:
                        print(f"\n-> {dtype} not recognised for {esc_name}.\n")
                        return
                    # checks
                    if d.size == 0:
                        # no data
                        if (
                            esc_name.split(":")[1].strip()
                            in self.allstrainsconditions[e]
                        ):
                            print(
                                f"\n-> No data found for {dtype} for {esc_name}.\n"
                            )
                        continue
                    # run fit
                    _, warning = runfitderiv(
                        self,
                        t,
                        d,
                        fitvar,
                        derivname,
                        e,
                        c,
                        s,
                        bd=bd,
                        cvfn=cvfn,
                        empirical_errors=empirical_errors,
                        noruns=noruns,
                        exitearly=exitearly,
                        noinits=noinits,
                        nosamples=nosamples,
                        logs=logs,
                        figs=figs,
                        findareas=findareas,
                        plotlocalmax=plotlocalmax,
                        showpeakproperties=showpeakproperties,
                        linalgmax=linalgmax,
                        max_data_pts=max_data_pts,
                        **kwargs,
                    )
                    if warning:
                        warnings += warning
    if warnings:
        print(warnings)



from pathlib import Path

import numpy as np
import pandas as pd

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omgenutils as gu


def savelog(self, fname=None):
    """
    Save log to file.

    Parameters
    --
    fname: string, optional
        The name of the file. If unspecified, the name of the experiment.

    Example
    -------
    >>> p.savelog()
    """
    # export log
    if fname:
        fnamepath = self.wdirpath / (fname + ".log")
    else:
        fnamepath = self.wdirpath / ("".join(self.allexperiments) + ".log")
    with fnamepath.open("w") as f:
        f.write(self.logstream.getvalue())
    print("Exported successfully.")


@clogger.log
def exportdf(self, fname=False, type="tsv", ldirec=False):
    """
    Export the dataframes.

    The exported data may either be tab-delimited or csv or json files.
    Dataframes for the (processed) raw data, for summary data, and for
    summary statistics and corrections, as well as a log file, are
    exported.

    Parameters
    ----------
    fname: string, optional
        The name used for the output files.
        If unspecified, the experiment or experiments is used.
    type: string
        The type of file for export, either 'json' or 'csv' or 'tsv'.
    ldirec: string, optional
        The directory to write. If False, the working directory is used.

    Examples
    --------
    >>> p.exportdf()
    >>> p.exportdf('processed', type= 'json')
    """
    if not fname:
        fname = "".join(self.allexperiments)
    if ldirec:
        ldirec = Path(ldirec)
        fullfname = str(ldirec / fname)
    else:
        fullfname = str(self.wdirpath / fname)
    # export data
    if type == "json":
        self.r.to_json(fullfname + "_r.json", orient="split")
        self.s.to_json(fullfname + "_s.json", orient="split")
        self.sc.to_json(fullfname + "_sc.json", orient="split")
    else:
        sep = "\t" if type == "tsv" else ","
        self.r.to_csv(fullfname + "_r." + type, sep=sep, index=False)
        self.s.to_csv(fullfname + "_s." + type, sep=sep, index=False)
        self.sc.to_csv(fullfname + "_sc." + type, sep=sep, index=False)
    # export log to file
    self.savelog(fname)


@clogger.log
def importdf(self, commonnames, info=True, sep="\t"):
    """
    Import dataframes saved as either json or csv or tsv files.

    Parameters
    ----------
    commonnames: list of strings
        A list of names for the files to be imported with one string for
        each experiment.

    Examples
    --------
    >>> p.importdf('Gal')
    >>> p.importdf(['Gal', 'Glu', 'Raf'])
    """
    commonnames = gu.makelist(commonnames)
    # import data (uses pandas but implicitly - there directly call pd.__version__)
    pd.__version__
    for commonname in commonnames:
        commonname = str(self.wdirpath / commonname)
        for df in ["r", "s", "sc"]:
            try:
                # json files
                exec(
                    "impdf= pd.read_json(commonname + '_' + df + "
                    "'.json', orient= 'split')"
                )
                print("Imported", commonname + "_" + df + ".json")
            except FileNotFoundError:
                try:
                    # csv files
                    exec(
                        "impdf= pd.read_csv(commonname + '_' + df + "
                        "'.csv', sep= ',')"
                    )
                    print("Imported", commonname + "_" + df + ".csv")
                except FileNotFoundError:
                    try:
                        # tsv files
                        exec(
                            "impdf= pd.read_csv(commonname + '_' + df "
                            "+ '.tsv', sep= '\t')"
                        )
                        print("Imported", commonname + "_" + df + ".tsv")
                    except FileNotFoundError:
                        print(
                            f"No file called {commonname}_{df}.json "
                            "or .csv or .tsv found"
                        )
                        return
            # ensure all are imported as strings
            for var in ["experiment", "condition", "strain"]:
                exec("impdf[var]= impdf[var].astype(str)")
            # merge dataframes
            if hasattr(self, df):
                exec(
                    "self."
                    + df
                    + "= pd.merge(self."
                    + df
                    + ", impdf, how= 'outer')"
                )
            else:
                exec("self." + df + "= impdf")
        print()

    # update attributes
    self.allexperiments = list(self.s.experiment.unique())
    self.allconditions.update(
        {
            e: list(self.s[self.s.experiment == e].condition.unique())
            for e in self.allexperiments
        }
    )
    self.allstrains.update(
        {
            e: list(self.s[self.s.experiment == e].strain.unique())
            for e in self.allexperiments
        }
    )
    for e in self.allexperiments:
        rdf = self.r[self.r.experiment == e]
        res = list((rdf.strain + " in " + rdf.condition).dropna().unique())
        res = [r for r in res if r != "nan in nan"]
        self.allstrainsconditions.update({e: res})

    # find datatypes with mean in self.s
    dtypdict = {}
    for e in self.allexperiments:
        # drop columns of NaNs - these are created by merge if a datatype
        # is in one experiment but not in another
        tdf = self.s[self.s.experiment == e].dropna(axis=1, how="all")
        dtypdict[e] = list(tdf.columns[tdf.columns.str.contains("mean")])
    self.datatypes.update(
        {e: [dt.split("_mean")[0] for dt in dtypdict[e]] for e in dtypdict}
    )
    # initialise progress
    for e in self.allexperiments:
        admin.initialiseprogress(self, e)
    # display info on import
    if info:
        self.info

    # display warning if duplicates created
    if len(self.allexperiments) != np.unique(self.allexperiments).size:
        print(
            "\nLikely ERROR: data with the same experiment, condition, "
            "strain, and time now appears twice!!"
        )


def save(self, name):
    """Remind user that save is undefined."""
    print("You probably mean exportdf.")



import numpy as np

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omerrors as errors
import omniplate.omgenutils as gu
import omniplate.sunder as sunder


def contentsofwells(self, wlist):
    """
    Display contents of wells.

    Parameters
    ----------
    wlist: string or list of string
        Specifies the well or wells of interest.

    Examples
    --------
    >>> p.contentsofwells(['A1', 'E4'])
    """
    wlist = gu.makelist(wlist)
    for w in wlist:
        print("\n" + w + "\n--")
        print(
            self.wellsdf.query("well == @w")
            .drop(["well"], axis=1)
            .to_string(index=False)
        )


def showwells(
    self,
    concise=False,
    sortby=False,
    experiments="all",
    conditions="all",
    strains="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditionincludes=False,
    conditionexcludes=False,
    strainincludes=False,
    strainexcludes=False,
):
    """
    Display wells for specified experiments, conditions, and strains.

    Parameters
    ----------
    concise: boolean
        If True, display as experiment: condition: strain.
    sortby: list of strings, optional
        List of column names on which to sort the results.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.

    Examples
    --------
    >>> p.showwells()
    >>> p.showwells(strains= 'Mal12:GFP', conditions= '1% Mal')
    """
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=False,
    )
    if not hasattr(self, "wellsdf"):
        self.wellsdf = admin.makewellsdf(self.r)
    df = self.wellsdf.query(
        "experiment == @exps and condition == @cons and strain == @strs"
    )
    if sortby:
        df = df.sort_values(by=gu.makelist(sortby))
    print()
    for e in exps:
        if concise:
            print(
                df[["experiment", "condition", "strain"]]
                .drop_duplicates()
                .query("experiment == @e")
                .to_string(index=False)
            )
        else:
            print(df.query("experiment == @e").to_string(index=False))
        print()


@clogger.log
def ignorewells(
    self,
    exclude=[],
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    clearall=False,
):
    """
    Ignore the wells specified in any future processing.

    If called several times, the default behaviour is for any previously
    ignored wells not to be re-instated.

    Parameters
    ---------
    exclude: list of strings
        List of labels of wells on the plate to be excluded.
    experiments: string or list of strings
        The experiments to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    clearall: boolean
        If True, all previously ignored wells are re-instated.

    Example
    -------
    >>> p.ignorewells(['A1', 'C2'])
    """
    if clearall:
        # forget any previously ignoredwells
        self.r = self.origr.copy()
        self.progress["ignoredwells"] = {exp: [] for exp in self.allexperiments}
        admin.update_s(self)
        print(
            "Warning: all corrections and analysis to raw data have been"
            " lost. No wells have been ignored."
        )
    else:
        if gu.islistempty(exclude):
            return
        else:
            # exclude should be a list of lists
            if isinstance(exclude, str):
                exclude = [gu.makelist(exclude)]
            elif isinstance(exclude[0], str):
                exclude = [exclude]
            # check consistency
            if len(self.allexperiments) == 1:
                exps = self.allexperiments
            else:
                exps = sunder.getset(
                    self,
                    experiments,
                    experimentincludes,
                    experimentexcludes,
                    "experiment",
                    nonull=True,
                )
            if len(exclude) != len(exps) and not clearall:
                raise errors.IgnoreWells(
                    "Either a list of wells to exclude for a particular\n"
                    "experiment or a list of experiments must be given."
                )
            else:
                # drop wells
                for ex, exp in zip(exclude, exps):
                    # wells cannot be ignored twice
                    wex = list(set(ex) - set(self.progress["ignoredwells"][exp]))
                    # drop data from ignored wells
                    df = self.r
                    filt = (df["experiment"] == exp) & df["well"].isin(wex)
                    df = df.loc[~filt]
                    df = df.reset_index(drop=True)
                    self.r = df
                    # store ignoredwells
                    self.progress["ignoredwells"][exp] += ex
                    # remove any duplicates
                    self.progress["ignoredwells"][exp] = list(
                        set(self.progress["ignoredwells"][exp])
                    )
                # give warning
                anycorrections = np.count_nonzero(
                    self.sc[[col for col in self.sc.columns if "correct" in col]].values
                )
                if np.any(anycorrections):
                    print(
                        "Warning: you have ignored wells after correcting\n"
                        "the data. It is best to ignorewells first, before\n"
                        "running any analysis."
                    )
            # remake summary data
            admin.update_s(self)



import matplotlib.pyplot as plt

import omniplate.admin as admin
import omniplate.omstats as omstats
from omniplate.omfitderiv import fitderiv


def runfitderiv(
    self,
    t,
    d,
    fitvar,
    derivname,
    experiment,
    condition,
    strain,
    bd=False,
    cvfn="matern",
    empirical_errors=False,
    noruns=10,
    exitearly=True,
    noinits=100,
    nosamples=100,
    logs=False,
    figs=True,
    findareas=False,
    plotlocalmax=True,
    showpeakproperties=False,
    linalgmax=5,
    max_data_pts=None,
    **kwargs,
):
    """
    Run fitderiv to smooth and estimate time derivatives for a single data set.

    Parameters
    ----------
    t: array
        An array of times.
    d: array
        An array of measurements of the variable to be fit.
    fitvar: string
        The name of the variable to be fit.
    derivname: string
        The name of the first time derivative of the variable.
    experiment: string
        The name of the experiment of interest.
    condition: string
        The condition of interest.
    strain: string
        The strain of interest.
    ylabels: list of strings
        The labels for the y-axis
    bd: dictionary, optional
        The bounds on the hyperparameters for the Gaussian process.
        For example, bd= {1: [-2,0])} fixes the bounds on the
        hyperparameter controlling flexibility to be 1e-2 and 1e0.
        The default for a Matern covariance function
        is {0: (-5,5), 1: (-4,4), 2: (-5,2)},
        where the first element controls amplitude, the second controls
        flexibility, and the third determines the magnitude of the
        measurement error.
    cvfn: string, optional
        The covariance function used in the Gaussian process, either
        'matern' or 'sqexp' or 'nn'.
    empirical_errors: boolean, optional
        If True, measurement errors are empirically estimated from the
        variance across replicates at each time point and so vary with
        time.
        If False, the magnitude of the measurement error is fit from the
        data assuming that this magnitude is the same at all time points.
    noruns: integer, optional
        The number of attempts made for each fit. Each attempt is made
        with random initial estimates of the hyperparameters within their
        bounds.
    exitearly: boolean, optional
        If True, stop at the first successful fit.
        If False, use the best fit from all successful fits.
    noinits: integer, optional
        The number of random attempts to find a good initial condition
        before running the optimization.
    nosamples: integer, optional
        The number of samples used to calculate errors in statistics by
        bootstrapping.
    logs: boolean, optional
        If True, find the derivative of the log of the data and should be
        True to determine the specific growth rate when dtype= 'OD'.
    figs: boolean, optional
        If True, plot both the fits and inferred derivative.
    findareas: boolean, optional
        If True, find the area under the plot of gr vs OD and the area
        under the plot of OD vs time. Setting to True can make getstats
        slow.
    plotlocalmax: boolean, optional
        If True, mark the highest local maxima found, which is used to
        calculate statistics, on any plots.
    showpeakproperties: boolean, optional
        If True, show properties of any local peaks that have found by
        scipy's find_peaks. Additional properties can be specified as
        kwargs and are passed to find_peaks.
    linalgmax: int, optional
        The number of linear algebra errors to tolerate.
    max_data_pts: integer, optional
        If set, sufficiently large data sets with multiple replicates will
        be subsampled at each time point, randomly picking a smaller
        number of replicates, to reduce the number of data points and so
        run times.
    """
    print(f"Fitting {fitvar} for {experiment}: {strain} in {condition}")
    # define statnames
    statnames = [
        f"min_{fitvar}",
        f"max_{fitvar}",
        f"range_{fitvar}",
        f"max_{derivname}",
        f"time_of_max_{derivname}",
    ]
    if derivname == "gr":
        # special names when estimating specific growth rate
        statnames += ["doubling_time", "lag_time"]
    else:
        statnames += [
            f"doubling_time_from_{derivname}",
            f"lag_time_from_{derivname}",
        ]
    # call fitderiv
    f = fitderiv(
        t,
        d,
        cvfn=cvfn,
        logs=logs,
        noruns=noruns,
        noinits=noinits,
        exitearly=exitearly,
        bd=bd,
        empirical_errors=empirical_errors,
        linalgmax=linalgmax,
        max_data_pts=max_data_pts,
    )
    if f.success:
        if figs:
            plt.figure()
            plt.subplot(2, 1, 1)
            f.plotfit(
                "f",
                ylabel=fitvar,
                figtitle=f"{experiment}: {strain} in {condition}",
            )
            axgr = plt.subplot(2, 1, 2)
            f.plotfit("df", ylabel=derivname)
            plt.tight_layout()
        else:
            axgr = None
        # find summary statistics
        (
            df_for_s,
            dict_for_sc,
            warning,
        ) = omstats.findsummarystats(
            fitvar,
            derivname,
            statnames,
            nosamples,
            f,
            t,
            experiment,
            condition,
            strain,
            findareas,
            figs,
            plotlocalmax,
            axgr,
            showpeakproperties,
            **kwargs,
        )
        # store GP parameters
        dict_for_sc[f"logmaxlikehood_for_{derivname}"] = f.logmaxlike
        dict_for_sc["gp_for_" + derivname] = cvfn
        for j, val in enumerate(f.lth):
            dict_for_sc[f"log_hyperparameter_{j}_for_{derivname}"] = val
        # add time series to s dataframe
        admin.add_to_s(self, derivname, df_for_s)
        # create or add summary stats to sc dataframe
        admin.add_dict_to_sc(self, dict_for_sc)
        if figs:
            plt.show(block=False)
        return f, warning
    else:
        return f, None
